﻿' (C) Copyright 2010 by Autodesk, Inc. 
'
' Permission to use, copy, modify, and distribute this software in
' object code form for any purpose and without fee is hereby granted,
' provided that the above copyright notice appears in all copies and
' that both that copyright notice and the limited warranty and
' restricted rights notice below appear in all supporting
' documentation.
'
' AUTODESK PROVIDES THIS PROGRAM "AS IS" AND WITH ALL FAULTS. 
' AUTODESK SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTY OF
' MERCHANTABILITY OR FITNESS FOR A PARTICULAR USE.  AUTODESK, INC. 
' DOES NOT WARRANT THAT THE OPERATION OF THE PROGRAM WILL BE
' UNINTERRUPTED OR ERROR FREE.
'
' Use, duplication, or disclosure by the U.S. Government is subject
' to restrictions set forth in FAR 52.227-19 (Commercial Computer
' Software - Restricted Rights) and DFAR 252.227-7013(c)(1)(ii)
' (Rights in Technical Data and Computer Software), as applicable.

Imports System
Imports Autodesk.AutoCAD.Runtime
Imports Autodesk.AutoCAD.ApplicationServices
Imports Autodesk.AutoCAD.DatabaseServices
Imports Autodesk.AutoCAD.Geometry
Imports Autodesk.AutoCAD.EditorInput
Imports System.Collections.Generic


Namespace TidyLines_VB


  Public Class MyCommands


    <CommandMethod("TIDYLINES")> _
    Public Shared Sub command(ByVal TidyLines)
      Dim db As Database = Application.DocumentManager.MdiActiveDocument.Database
      Dim ed As Editor = Application.DocumentManager.MdiActiveDocument.Editor

      ' Prompt user to select set of lines
      Dim opts As New PromptSelectionOptions
      opts.AllowDuplicates = False
      opts.AllowSubSelections = False
      opts.MessageForAdding = "Select Lines :"
      opts.MessageForRemoval = "Remove Lines :"
      Dim values() As TypedValue = {New TypedValue(DxfCode.Start, "LINE")}
      Dim sfilter As New SelectionFilter(values) ' Create the filter using our values...

      Dim res As PromptSelectionResult = ed.GetSelection(opts, sfilter)
      If res.Status <> PromptStatus.OK Then Exit Sub

      Dim ids() As ObjectId = res.Value.GetObjectIds
      If ids.Length = 0 Then
        ed.WriteMessage(vbLf & "No lines were selected." & vbLf)
        Exit Sub
      End If

      ' Prompt user to enter endpoint tolerance 'n'
      Dim opts2 As New PromptIntegerOptions("Specify endpoint tolerance % (1-49)")
      opts2.LowerLimit = 1
      opts2.UpperLimit = 49
      Dim res2 As PromptIntegerResult = ed.GetInteger(opts2)
      If res2.Status <> PromptStatus.OK Then Exit Sub
      Dim n As Integer = res2.Value

      ' Prompt user to enter perpendicular tolerance
      Dim opts3 As New PromptIntegerOptions("Specify perpendicular tolerance % (1-10)")
      opts3.LowerLimit = 1
      opts3.UpperLimit = 10
      Dim res3 As PromptIntegerResult = ed.GetInteger(opts3)
      If res3.Status <> PromptStatus.OK Then Exit Sub
      Dim p As Integer = res3.Value

      Using tr As Transaction = db.TransactionManager.StartTransaction
        'Open all the lines for read ready for processing
        Dim coll As New DBObjectCollection
        For Each objId As ObjectId In ids
          coll.Add(tr.GetObject(objId, OpenMode.ForRead))
        Next

        'Setup some variables we'll need in the next loop
        Dim coincidentColl As New List(Of ConstraintData)
        Dim iPtr As New IntPtr(0)
        Dim pts As New Point3dCollection

        ' For each line, iterate through all the other lines, finding the lines that intersect it within 'n'% of 
        '  its length from the start or end point (e.g. 25 means intersection point is within 25% if line length of one end).
        ' Setup loop so we don't try to self intersect a line, or intersect line B with line A when we've already intersected A with B.
        For i As Integer = 0 To coll.Count - 2
          Dim line1 As Line = coll.Item(i)
          For j As Integer = i + 1 To coll.Count - 1
            Dim line2 As Line = coll.Item(j)
            ' Have to check how close intersection is to the end points of both lines.
            ' Possible hits are start-start, end-end, start-end, end-start

            ' Intersect the lines
            pts.Clear()
            line1.IntersectWith(line2, Intersect.OnBothOperands, pts, iPtr, iPtr)
            ' These are lines, so we only expect a single intersection point (at most)
            ' End this loop if we get more points than that
            If pts.Count <> 1 Then Continue For

            ' Find distance of intersection point along both lines
            ' dist 1 is distance of intersection along line 1, dist2 is same for line2
            Dim dist1 As Double = line1.GetDistAtPoint(pts(0))
            Dim dist2 As Double = line2.GetDistAtPoint(pts(0))


            ' Now check if its within 'n'% of either end
            Dim subEnt1 As FullSubentityPath = CloseToEnd(line1, dist1, n)
            If subEnt1 = FullSubentityPath.Null Then Continue For
            Dim subEnt2 As FullSubentityPath = CloseToEnd(line2, dist2, n)
            If subEnt2 = FullSubentityPath.Null Then Continue For

            ' Now check if lines are close to perpendicular
            Dim angle As Double = line1.Delta.GetAngleTo(line2.Delta)
            ' Calculate tolerance angle from tolerance value p entered by user
            ' So if angle is within p hundredths of 90 degrees, we assume it is supposed to be 90 degrees.
            Dim tolAngle As Double = p * (Math.PI / 2) / 100
            Dim bPerp As Boolean
            If (angle < (Math.PI / 2) + tolAngle) And (angle > (Math.PI / 2) - tolAngle) Then
              bPerp = True
            End If

            ' If we get here, then we have a match on both lines - subEnt1 and subEnt2 are valid FullSubentityPaths
            ' Now we add constraint data so we can create the constraints after we finished looping
            coincidentColl.Add(New ConstraintData(getSubEnt(line1, SubEntInfo.Edge), getSubEnt(line2, SubEntInfo.Edge), subEnt1, subEnt2, bPerp))


          Next
        Next

        ' For lines that match, create coincident constraint.
        If coincidentColl.Count > 0 Then
          Dim consGrpId As ObjectId = GetConstraintGroup(True)
          Dim consGrp As Assoc2dConstraintGroup = tr.GetObject(consGrpId, OpenMode.ForWrite)
          Dim bFlag As Boolean
          For Each consData In coincidentColl
            addConstrainedGeometry(consGrp, consData.Edge1)
            addConstrainedGeometry(consGrp, consData.Edge2)
            Dim paths As FullSubentityPath() = New FullSubentityPath(1) {consData.Vertex1, consData.Vertex2}
            Try
              consGrp.AddGeometricalConstraint(GeometricalConstraint.ConstraintType.Coincident, paths)
              If consData.Perpendicular Then
                consGrp.AddGeometricalConstraint(GeometricalConstraint.ConstraintType.Perpendicular, New FullSubentityPath(1) {consData.Edge1, consData.Edge2})
              End If
            Catch ex As Autodesk.AutoCAD.Runtime.Exception
              ' We're assuming that if an exception is thrown here, then its not fatal. Therefore, we eat it and carry on.
              ' If the error is anything other than AlreadyInGroup (which is a non-event), we warn the user later.
              If ex.ErrorStatus <> ErrorStatus.AlreadyInGroup Then
                bFlag = True
              End If
            End Try
          Next

          ' Its possible that running this command will over constrain the drawing.
          If bFlag Then
            ed.WriteMessage(vbLf & "Warning: Some calculated constraints could not be added. Processing the selected geometry was probably over constraining the drawing. " & vbLf)
          End If
        End If

        tr.Commit()
      End Using
    End Sub



    ' Helper class to hold data used to add constraints after we've analyzed all our lines
    Private Class ConstraintData
      Private mEdgeSE1 As FullSubentityPath
      Private mEdgeSE2 As FullSubentityPath
      Private mVertexSE1 As FullSubentityPath
      Private mVertexSE2 As FullSubentityPath
      Private mPerpendicular As Boolean
      Public Property Edge1() As FullSubentityPath
        Get
          Return mEdgeSE1
        End Get
        Set(ByVal value As FullSubentityPath)
          mEdgeSE1 = value
        End Set
      End Property
      Public Property Edge2() As FullSubentityPath
        Get
          Return mEdgeSE2
        End Get
        Set(ByVal value As FullSubentityPath)
          mEdgeSE2 = value
        End Set
      End Property
      Public Property Vertex1() As FullSubentityPath
        Get
          Return mVertexSE1
        End Get
        Set(ByVal value As FullSubentityPath)
          mVertexSE1 = value
        End Set
      End Property

      Public Property Vertex2() As FullSubentityPath
        Get
          Return mVertexSE2
        End Get
        Set(ByVal value As FullSubentityPath)
          mVertexSE2 = value
        End Set
      End Property

      Public Property Perpendicular() As Boolean
        Get
          Return mPerpendicular
        End Get
        Set(ByVal value As Boolean)
          mPerpendicular = value
        End Set
      End Property

      Public Sub New()
      End Sub

      Public Sub New(ByVal edge1 As FullSubentityPath, ByVal edge2 As FullSubentityPath, ByVal vertex1 As FullSubentityPath, ByVal vertex2 As FullSubentityPath, ByVal perp As Boolean)
        Me.Edge1 = edge1
        Me.Edge2 = edge2
        Me.Vertex1 = vertex1
        Me.Vertex2 = vertex2
        Me.Perpendicular = perp
      End Sub
    End Class



    ' Passed to getSubEnt function to specify what subentity of line we want to retrieve
    Private Enum SubEntInfo
      StartPoint
      EndPoint
      MidPoint
      Edge
    End Enum


    ' Returns FullSubEntPath of vertex if intersection point was closer to that vertex than the specified parameter.
    ' E.g. If param is 25, we'll return the startpoint vertex if dist is less than 25% of the length of the line,
    '  the endpoint vertex if dist is more than 75% along the line,
    '  or null FullSubentityPatgh if it was between 25 and 75%
    Private Shared Function CloseToEnd(ByVal lin As Line, ByVal dist As Double, ByVal param As Integer) As FullSubentityPath

      If 100 * dist / lin.Length < param Then
        ' Close to start
        Return getSubEnt(lin, SubEntInfo.StartPoint)
      ElseIf 100 * (lin.Length - dist) / lin.Length < param Then
        ' Close to end
        Return getSubEnt(lin, SubEntInfo.EndPoint)
      Else
        ' Not close to start or end
        Return FullSubentityPath.Null
      End If

    End Function





    ' Helper function to return the FullSubEntityPath for the required subent type for the entity passed in.
    ' Sub ent options are edge, srartpoint, midpoint, endpoint.
    Private Shared Function getSubEnt(ByVal ent As Entity, ByVal info As SubEntInfo) As FullSubentityPath
      ' Instantiate protocol extension required to query entity
      Dim subentityIdPE As AssocPersSubentityIdPE
      Dim peCls As RXClass = AssocPersSubentityIdPE.GetClass(GetType(AssocPersSubentityIdPE))
      Dim pSubentityIdPE As IntPtr = ent.QueryX(peCls)
      If pSubentityIdPE = IntPtr.Zero Then
        System.Windows.MessageBox.Show("cannot get pSubentityIdPE")
        Return FullSubentityPath.Null
      End If
      subentityIdPE = TryCast(AssocPersSubentityIdPE.Create(pSubentityIdPE, False), AssocPersSubentityIdPE)
      If subentityIdPE Is Nothing Then
        System.Windows.MessageBox.Show("cannot get subentityIdPE")
        Return FullSubentityPath.Null
      End If
      Dim edgeSubentityIds As SubentityId() = Nothing
      edgeSubentityIds = subentityIdPE.GetAllSubentities(ent, SubentityType.Edge)
      ' Assume we want first edge in returned array
      Dim subentEdgePath1 As New FullSubentityPath(New ObjectId(0) {ent.ObjectId}, edgeSubentityIds(0))
      If info = SubEntInfo.Edge Then
        Return subentEdgePath1
      End If

      ' Edge wasn't wanted, so return the specified vertex.
      Dim start As SubentityId = SubentityId.Null, [end] As SubentityId = SubentityId.Null
      Dim other As SubentityId() = Nothing
      subentityIdPE.GetEdgeVertexSubentities(ent, edgeSubentityIds(0), start, [end], other)
      If info = SubEntInfo.StartPoint Then
        Return New FullSubentityPath(New ObjectId(0) {ent.ObjectId}, start)
      End If
      If info = SubEntInfo.EndPoint Then
        Return New FullSubentityPath(New ObjectId(0) {ent.ObjectId}, [end])
      End If
      Return FullSubentityPath.Null
    End Function




    '  Helper function to retrieve (or create) constraint group
    Private Shared Function GetConstraintGroup(ByVal createIfDoesNotExist As Boolean) As ObjectId
      ' Calculate the current plane on which new entities are added by the editor
      ' (A combination of UCS and ELEVATION sysvar).
      Dim ed As Autodesk.AutoCAD.EditorInput.Editor = Autodesk.AutoCAD.ApplicationServices.Application.DocumentManager.MdiActiveDocument.Editor
      Dim ucsMatrix As Matrix3d = ed.CurrentUserCoordinateSystem
      Dim origin As Point3d = ucsMatrix.CoordinateSystem3d.Origin
      Dim xAxis As Vector3d = ucsMatrix.CoordinateSystem3d.Xaxis
      Dim yAxis As Vector3d = ucsMatrix.CoordinateSystem3d.Yaxis
      Dim zAxis As Vector3d = ucsMatrix.CoordinateSystem3d.Zaxis
      origin = origin + CDbl(Application.GetSystemVariable("ELEVATION")) * zAxis
      Dim currentPlane As New Plane(origin, xAxis, yAxis)

      ' get the constraint group from block table record
      Dim idConstrGroup As ObjectId = ObjectId.Null
      Dim db As Database = Application.DocumentManager.MdiActiveDocument.Database
      Dim networkId As ObjectId = AssocNetwork.GetInstanceFromObject(db.CurrentSpaceId, createIfDoesNotExist, True, "")
      If networkId.IsNull Then
        System.Windows.MessageBox.Show("network id is null")
        Return idConstrGroup
      End If

      ' Try to find the constraint group in the associative network
      Dim tm As Autodesk.AutoCAD.DatabaseServices.TransactionManager = db.TransactionManager
      Using myT As Transaction = tm.StartTransaction()
        Using network As AssocNetwork = DirectCast(myT.GetObject(networkId, OpenMode.ForRead, False), AssocNetwork)
          If network Is Nothing Then
            Return idConstrGroup
          End If
          ' Iterate all actions in network to find Assoc2dConstraintGroups
          Dim actionsInNetwork As ObjectIdCollection = network.GetActions
          For nCount As Integer = 0 To actionsInNetwork.Count - 1
            Dim idAction As ObjectId = actionsInNetwork(nCount)
            If idAction = ObjectId.Null Then
              Continue For
            End If
            ' Is this action a type of Assoc2dConstraintGroup?
            If idAction.ObjectClass.IsDerivedFrom(RXObject.GetClass(GetType(Autodesk.AutoCAD.DatabaseServices.Assoc2dConstraintGroup))) Then
              Using action As AssocAction = DirectCast(myT.GetObject(idAction, OpenMode.ForRead, False), AssocAction)
                If action Is Nothing Then
                  Continue For
                End If
                Dim constGrp As Assoc2dConstraintGroup = DirectCast(action, Assoc2dConstraintGroup)
                ' Is this the Assoc2dConstraintGroup for our plane of interest?
                If constGrp.GetWorkPlane.IsCoplanarTo(currentPlane) Then
                  ' If it is then we've found an existing constraint group we can use.
                  Return idAction
                End If
              End Using
            End If
          Next
        End Using

        ' If we get to here, a suitable contraint group doesn't exist, create a new one if that's what calling fn wanted.
        If idConstrGroup.IsNull AndAlso createIfDoesNotExist Then
          Using network As AssocNetwork = DirectCast(myT.GetObject(networkId, OpenMode.ForWrite, False), AssocNetwork)
            ' Create  construction plane
            Dim constraintPlane As New Plane(currentPlane)
            ' If model extent is far far away from origin then we need to shift 
            '  construction plane origin within the model extent.
            ' (Use Pextmin, PExtmax in paper space)
            Dim bIsModelSpace As Boolean = db.TileMode
            Dim extmin As Point3d = If(bIsModelSpace, db.Extmin, db.Pextmin)
            Dim extmax As Point3d = If(bIsModelSpace, db.Extmax, db.Pextmax)
            If extmin.GetAsVector().Length > 100000000.0 Then
              Dim originL As Point3d = extmin + (extmax - extmin) / 2.0
              Dim result As PointOnSurface = currentPlane.GetClosestPointTo(originL)
              constraintPlane.[Set](result.GetPoint(), currentPlane.Normal)
            End If
            ' Create the new constraint group and add it to the associative network.
            Using constGrp As New Assoc2dConstraintGroup(constraintPlane)
              idConstrGroup = db.AddDBObject(constGrp)
            End Using
            network.AddAction(idConstrGroup, True)
          End Using
        End If
        myT.Commit()
      End Using
      Return idConstrGroup
    End Function



    ' Helper function to handle that an exception is thrown if you try to add the same constrained geometry twice.
    ' You could work out what constrained geometry to add up front. That would be more efficient as no exceptions would be thrown,
    '  but isn't always possible (e.g. if running a command multiple times to constrain existing geometry in a drawing
    Private Shared Function addConstrainedGeometry(ByVal constGrp As Assoc2dConstraintGroup, ByVal path As FullSubentityPath) As ConstrainedGeometry
      Dim consGeom As ConstrainedGeometry = Nothing
      Try
        consGeom = constGrp.AddConstrainedGeometry(path)
      Catch ex As Autodesk.AutoCAD.Runtime.Exception
        ' This error isn't really an an error
        If ex.ErrorStatus <> ErrorStatus.AlreadyInGroup Then
          Throw ex
        End If
        consGeom = constGrp.GetConstrainedGeometry(path, False)
      End Try
      Return consGeom
    End Function

  End Class

End Namespace